/*
 * filterdiff - extract (or exclude) a diff from a diff file
 * lsdiff - show which files are modified by a patch
 * Copyright (C) 2001 Tim Waugh <twaugh@redhat.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <fnmatch.h>
#include <getopt.h>
#include <locale.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifdef HAVE_SYS_TYPES_H
# include <sys/types.h> // for ssize_t
#endif /* HAVE_SYS_TYPES_H */
#include <time.h>

#include "util.h"
#include "diff.h"

static struct patlist *pat_include = NULL;
static struct patlist *pat_exclude = NULL;

static int unzip = 0;
static int listing;
static int numbering = 0;
static int ignore_components = 0;
static int show_status = 0;

static int file_exists (const char *timestamp)
{
	struct tm t;
	time_t s;

	if (!strptime (timestamp, "%a %b %e %T %Y", &t))
		return 1;

	s = mktime (&t);

	/* If s is negative, or less that fifteen hours into 1970,
	 * and it's an exact multiple of 15 minutes, it's very likely
	 * to be the result of ctime(&zero). */
	if (s < (15 * 60 * 60) && (s % (15 * 60)) == 0)
		return 0;

	/* Otherwise, it's a real file timestamp. */
	return 1;
}

static const char *stripped (const char *name)
{
	const char *basename;
	int i = 0;
	if (!strncmp (name, "./", 2))
		name += 2;
	basename = strrchr (name, '/');
	if (!basename)
		basename = name;
	while (i < ignore_components &&
	       (name = strchr (name, '/')) != NULL) {
		while (*name == '/')
			name++;
		i++;
	}
	return name ? name : basename;
}

static int filterdiff (FILE *f)
{
	unsigned long orig_lines, new_lines;
	static unsigned long linenum = 0;
	char *names[2];
	char *origline = NULL;
	char *line = NULL;
	size_t linelen = 0;
	char *p;
	int match;

	for (;;) {
		int h;
		int orig_file_exists;

		orig_file_exists = 0; // shut gcc up

		// Search for start of patch ("--- ").
		for (;;) {
			if (getline (&line, &linelen, f) == -1)
				goto eof;
			linenum++;

		check_orig:
			if (!strncmp (line, "--- ", 4))
				break;
		}

		origline = xstrdup (line);
		h = strcspn (line + 4, " \t\n");
		names[0] = xstrndup (line + 4, h);
		h = 4 + h + strspn (line + 4 + h, " \t\n");
		if (listing && show_status)
			orig_file_exists = file_exists (line + h);

		if (getline (&line, &linelen, f) == -1) {
			free (names[0]);
			goto eof;
		}
		linenum++;

		if (strncmp (line, "+++ ", 4)) {
			free (names[0]);
			goto check_orig;
		}

		h = strcspn (line + 4, " \t\n");
		names[1] = xstrndup (line + 4, h);

		// Decide whether this matches this pattern.
		p = best_name (2, names);
		
		match = !patlist_match(pat_exclude, p);
		if (match && pat_include != NULL)
			match = patlist_match(pat_include, p);
		
		// print if it matches.
		if (match) {
			if (listing) {
				if (numbering)
					printf ("%6lu  ", linenum - 1);
				if (show_status) {
					char st = '!';
					h = 4 + h + strspn (line + 4 + h,
							    " \t\n");
					if (!orig_file_exists)
						st = '+';
					else if (!file_exists (line + h))
						st = '-';
					printf ("%c ", st);
				}
				puts (stripped (p));
			} else {
				fputs (origline, stdout);
				fputs (line, stdout);
			}
		}

		free (names[0]);
		free (names[1]);
		free (origline);
		origline = NULL;

		if (getline (&line, &linelen, f) == -1)
			goto eof;
		linenum++;

		if (strncmp (line, "@@ ", 3))
			goto check_orig;

		orig_lines = orig_num_lines (line);
		new_lines = new_num_lines (line);

		if (match && !listing)
			fputs (line, stdout);

		for (;;) {
			if (getline (&line, &linelen, f) == -1)
				goto eof;
			linenum++;

			if (!orig_lines && !new_lines && line[0] != '\\') {
				if (strncmp (line, "@@ ", 3))
					break;

				/* Next chunk. */
				orig_lines = orig_num_lines (line);
				new_lines = new_num_lines (line);

				if (match && !listing)
					fputs (line, stdout);

				continue;
			}

			if (orig_lines && line[0] != '+')
				orig_lines--;
			if (new_lines && line[0] != '-')
				new_lines--;

			if (match && !listing)
				fputs (line, stdout);
		}

		goto check_orig;
	}

 eof:
	if (line)
		free (line);

	if (origline)
		free (origline);

	return 0;
}

static char * syntax_str =
"usage: %s [OPTION]... [files ...]\n"
"Options:\n"
"  -x PAT    exclude files matching PAT\n"
"  -i PAT    include only files matching PAT\n"
"  -z        decompress .gz and .bz2 files\n"
"  -n        show line numbers (lsdiff)\n"
"  -p N      pathname components to ignore (lsdiff)\n"
"  -s        show file additions and removals (lsdiff)\n"
"  --filter  run as 'filterdiff'\n"
"  --list    run as 'lsdiff'\n"
;

NORETURN
static void syntax (int err)
{
	fprintf (err ? stderr : stdout, syntax_str, progname);
	exit (err);
}

static void set_list (void)
{
	/* This is lsdiff. */
	set_progname ("lsdiff");
	listing = 1;
}

static void set_filter (void)
{
	/* This is filterdiff. */
	set_progname ("filterdiff");
	listing = 0;
}

static void list_or_filter (const char *argv0)
{
	/* This is filterdiff, unless it is named 'lsdiff'. */
	const char *p = strrchr (argv0, '/');
	if (!p++)
		p = argv0;
	if (strstr (p, "lsdiff"))
		set_list ();
	else
		set_filter ();
}

int main (int argc, char *argv[])
{
	int i;
	FILE *f = stdin;

	setlocale (LC_TIME, "C");
	list_or_filter (argv[0]);
	while (1) {
		static struct option long_options[] = {
	       		{"help", 0, 0, 'h'},
			{"version", 0, 0, 'v'},
			{"list", 0, 0, 'l'},
			{"filter", 0, 0, 'f'},
			{0, 0, 0, 0}
		};
		char *end;
		int c = getopt_long (argc, argv, "vhp:i:I:x:X:zns",
				long_options, NULL);
		if (c == -1)
			break;
		
		switch (c) {
		case 'f':
			set_filter ();
			break;
		case 'l':
			set_list ();
			break;
		case 'v':
			printf("%s - patchutils version %s\n", progname,
			       VERSION);
			exit(0);
		case 'h':
			syntax (0);
			break;
		case 'p':
			ignore_components = strtoul (optarg, &end, 0);
			if (optarg == end)
				syntax (1);
			break;
		case 'x':
			patlist_add (&pat_exclude, optarg);
			break;
		case 'i':
			patlist_add (&pat_include, optarg);
			break;
		case 'z':
			unzip = 1;
			break;
		case 'n':
			numbering = 1;
			break;
		case 's':
			show_status = 1;
			break;
		default:
			syntax(1);
		}
				
	}
	
	if (optind == argc) {
		filterdiff(stdin);
	} else {
		for (i = optind; i < argc; i++) {
			if (unzip) {
				f = xopen_unzip (argv[i], "rb");
			} else {
				f = xopen(argv[i], "rb");
			}
			filterdiff(f);
			fclose(f);
		}
	}

	return 0;
}

